!===================================================================
subroutine cut_off()
use param
use mesh
use elements 
use ice
use parsup
implicit none

where(a_ice>1.0_8)
 a_ice=1.0_8
end where

where(a_ice<0.1e-8)
 a_ice=0.0_8
end where

where(m_ice<0.1e-8)
 m_ice=0.0_8
end where

end subroutine cut_off
! ==================================================================
subroutine thermodynamics
!
! For every surface node, this subroutine extracts the information
! needed for computation of thermodydnamics, calls the relevant
! subroutine, and returns the information to the vectors of prognostic
! variables.
!
! Originally programmed by N. Yakovlev/S. Danilov.
! Adjusted for upgraded model physics (NCEP forcing data; parameterization
! of ice-ocean heat flux considering friction velocity) by Ralph Timmermann.
! Note that adjustments are not indivually labeled any more.
!===================================================================
use param
use mesh
use elements 
use ice
use parsup
use ice_therm_parms
use ice_dyn_parms
USE ATM_FORCING

implicit none
real*8    h,hsn,A,t,fsh,flo,Ta,td,qa,preslev,acl,precrate
real*8    ug,ustar,T_oc,S_oc,fw,ehf,h_ml
integer i, j


do j=1, myDim_nod2D+eDim_nod2D
        i       = myList_nod2D(j)
        h       = m_ice(i)
        hsn     = m_snow(i)
        A       = a_ice(i)
        t       = 0.                   ! output variable
        fsh     = shortwave(i)
        flo     = longwave(i)
        Ta      = Tair(i)
        td      = Tdew(i)
        qa      = shum(i) 
        acl     = cloudiness(i)
        preslev = Pair(i)       
        precrate= precipitation(i)
        ug      = sqrt(u_wind(i)**2+v_wind(i)**2)
        ustar   = sqrt(cdwat)*sqrt((u_ice(i)-u_w(i))**2+(v_ice(i)-v_w(i))**2)
        T_oc    = T_oc_array(i)      
        S_oc    = S_oc_array(i)
        h_ml    = 10.
        fw      = 0
        ehf     = 0.

        call therm_ice(h,hsn,A,t,fsh,flo,Ta,td,qa,acl,preslev,precrate, &
                       ug,ustar,T_oc,S_oc,h_ml,dt,i,fw,ehf)

        m_ice(i)         = h
        m_snow(i)        = hsn
        a_ice(i)         = A
        fresh_wa_flux(i) = fw
        net_heat_flux(i) = ehf
end do

end subroutine thermodynamics
!===================================================================

!====================================================================
subroutine therm_ice(h,hsn,A,t,fsh,flo,Tair,td,qa,acl,preslev,precrate, &
                     ug,ustar,T_oc,S_oc,H_ML,dt,inode,fw,ehf)

!===========================================================
!             Ice Thermodynamic growth model               !
!===========================================================

! Input parameters:
!------------------
! h - ice mass [m]
! hsn - snow mass [m]
! A - ice compactness
! t - temperature of snow/ice top surface
! fsh - shortwave radiation
! flo - longwave radiation
! tair - air temperature
! td - air dew point 
! preslev - air pressure [hPa]
! precrate - precipitation rate
! ug - wind speed
! T_oc, S_oc - ocean temperature and salinity beneath the ice (mixed layer)
! H_ML - mixed layer depth - should be specified.
! dt - time step
! ustar - friction velocity

! Output parameters:
!-------------------
! h - ice mass
! hsn - snow mass
! A - ice compactness
! t - temperature of snow/ice top surface
! prec - fresh water=precipitation+snow melt, going to ocean [m water]
! fw - freshwater flux due to ice melting [m water/dt]
! ehf - net heat flux at the ocean surface [W/m2]        !RTnew


use ice_therm_parms
use parsup

implicit none
real*8 h,hsn,A,t,dhgrowth,dhsngrowth,ahf,Tair,td,qa,acl,preslev,prec,precrate,  &      
	   fsh,flo,fh,gh,fw,ehf,ug,ustar,hfsen,hfsenow,hfrad,hfradow,hflatow,hftotow, & 
	   hflat,hftot,T_oc,S_oc,dt, H_ML, tmp,tmp2,tmp3

real*8 rhow,show,rhice,shice,sh,thick,thact
real*8 thsn                                     
real*8 rh,ra,qhst,qfm,qtm,sn,snow,hsntmp,qhsto,qt

!RT:
real*8 o2ihf
!RT-

integer K
integer inode

real*8, external :: TFrez  ! Sea water freeze temperature.
!real*8, external :: QToc   ! Ocean-Ice heat flux.  !rt: not used any more
   	  
!------------------------------------------------------------------
! determine h(i,j)/a(i,j) = actual ice thickness.
! if snow layer is present, add hsn weighted with quotient
! of conductivities of ice and snow, according to 0-layer approach
! of Semtner (1976).
!------------------------------------------------------------------

dhgrowth=h  ! Store ice thickness at start of growth routine
	    	    
thsn=hsn*(con/consn)/max(A,Armin)    ! Effective snow thickness
thick=h/max(A,Armin)+thsn            ! Effective ice thickness
!!!!!!!thick=max(thick,hmin)  ! Now there is IF block 28.10.2003.
rhice=0.0                            ! Reset average heat flux

!write(*,*) T_oc vor obudget,T_oc
call obudget (preslev,td,qa,fsh,flo,T_oc,ug,tair,rhow)  ! Growth rate for thin ice (open ocean)
!ine obudget (preslev,td,qa,fsh,flo,t,ug,tair,fh)  

if (thick.GT.hmin) then
! K loop for all ice classes
! write(mype+30,*) 'iclasses',istep,inode,iclasses,thick
 do K=1,iclasses			               ! K loop
  thact = (2*k-1)*thick/float(iclasses)  ! Thicknesses of actual ice class
!  write(mype+30,*) 'in iclasses',&
!   thact,hsn,t,tair,td,qa,preslev,fsh,flo,ug,S_oc,shice
  call budget (thact,hsn,t,tair,td,qa,preslev,fsh,flo,ug,S_oc,shice) ! Thick ice K-class growth rate
  rhice=rhice+shice/float(iclasses)      ! Add to average heat flux
 end do			                       ! k loop
END IF
!write(mype+30,*) 'after iclasses',istep,inode,iclasses,rhice

!---------------------------------------------------------------
! Convert growth rates [m ice/sec] into growth per time step DT.
!---------------------------------------------------------------

rhow=rhow*dt
rA=rhow		            ! for historical reasons
rhice=rhice*dt

!---------------------------------------------------------------
! Multiply ice growth of open water and ice
! with the corresponding areal fractions of grid cell
!---------------------------------------------------------------

show =rhow *(1.0-A)
shice=rhice*A
sh   =show+shice

ahf=-cl*sh/dt    ! Atmospheric heat flux, average over grid cell [W/m**2]

if (Tair.ge.0.) then ! Precipitation going to ocean
	prec=precrate*dt
	snow=0.
else
	prec=precrate*dt*(1.-A)
	snow=precrate*dt*A*rhowat/rhosno
endif

hsn=hsn+snow     ! Add snow fall to temporary snow thickness
dhsngrowth=hsn   ! Store snow thickness after snow fall 

! ---------------------------------------------------------------------
! If there is atmospheric melting, first melt any snow that is present.
! Atmospheric heat flux available for melting
! sh = MINUS atm. heat flux / specific latent heat of sea ice
! Note: (sh<0) for melting, (sh>0) for freezing
!----------------------------------------------------------------------
hsntmp= -min(sh,0.)*rhoice/rhosno

!--------------------------------------------------------------------
! hsntmp is the decrease in snow thickness due to atmospheric melting
! [m/DT]. Do not melt more snow than available
!--------------------------------------------------------------------
hsntmp=min(hsntmp,hsn)
hsn=hsn-hsntmp  ! Update snow thickness after atmospheric snow melt

!-------------------------------------------------------------------
! Negative atmospheric heat flux left after melting of snow
! Note: (sh<0) and (hsntmp>0) for melting conditions
! hsntmp=0 for non-snow-melting conditions
! RT-FR099: snow melt will be added to fw flux at the end of this routine
!-------------------------------------------------------------------
rh=sh+hsntmp*rhosno/rhoice
h=max(h,0.)

!-------------------------------------------------------------------
! Compute heat content qhst of mixed layer - sea ice system
!
! Total heat content is the sum of
!	h	ice thickness after calculation of dynamic effects
!	rh	change in ice thickness due to atmospheric forcing
! and heat available in mixed layer, with
!	T_oc	temperature of ocean surface layer
!	Tfrez	freezing point of sea water
!	H_ML	thickness of uppermost layer
!
!RT:
! There are three possibilities to do this.
! 1.: Assume an instantaneous adjustment of mixed layer heat content.
!     Any heat available is then instantaneously used to melt ice.
!     (so-called ice-bath approach)
!     This is what used to be used in the Lemke sea ice-mixed layer model.
! qhst=h+rh -(T_oc-TFrez(S_oc))*H_ML*cc/cl
!
! 2.: Parameterize the ocean-to-ice heat flux (o2ihf)
!     as a function of temperature difference. For a first step 
!     we can assume a constant exchange coefficient gamma_t:
! o2ihf= (T_oc-TFrez(S_oc))*gamma_t*cc*A     &
!        +(T_oc-Tfrez(S_oc))*H_ML/dt*cc*(1-A)  ! [W/m2]
! qhst=h+rh-o2ihf*dt/cl                        ! [m]
!
! 3.  Parameterize the ocean-to-ice heat flux (o2ihf)
!     as a function of temperature difference and the
!     friction velocity:
o2ihf= (T_oc-TFrez(S_oc))*0.012*ustar*cc*A  &
      +(T_oc-Tfrez(S_oc))*H_ML/dt*cc    *(1-A)  ! [W/m2]
qhst=h+rh-o2ihf*dt/cl                           ! [m]
!    
!RT--
!-------------------------------------------------------------------
! qhst > 0	temporary ice thickness [m]
! qhst < 0  	temporary heat stored in mixed layer (all ice melts)
!-------------------------------------------------------------------

!-------------------------------------------------------------------
! New temporary ice thickness after changes due to atmospheric
!  forcing and mixed layer heat content:
!-------------------------------------------------------------------
tmp=max(qhst,0.)
!-------------------------------------------------------------------
! Melt snow if there is any ML heat content left (qhst<0).
! This may be the case if advection moves ice (with snow) to regions
! with a warm mixed layer.
!-------------------------------------------------------------------
sn=hsn+min(qhst,0.)*rhoice/rhosno
!-------------------------------------------------------------------
! When the temporary snow thickness is negative, there is some ML
! heat content left: 
!-------------------------------------------------------------------
tmp2=min(sn,0.)*rhosno/rhoice
!-------------------------------------------------------------------
! tmp2 <= 0 is the ice melt equivalent [m] of the mixed layer heat
! content left after melting of snow.
! New temporary snow thickness must not be negative:
!-------------------------------------------------------------------
sn=max(sn,0.)
!-------------------------------------------------------------------
! Add snow melt to precipitation
!-------------------------------------------------------------------
!rt FR099: this will be done at the end of this routine
!rt prec=prec +(hsn-sn)*rhosno/rhowat
!-------------------------------------------------------------------
! "Add" snow melt to ML heat content (there will be no snow melt for qhst>0)
!-------------------------------------------------------------------
qhst=qhst+(hsn-sn)*rhosno/rhoice
!-------------------------------------------------------------------
! Additional fresh water flux from ice thickness changes
! due to atmospheric forcing and ML heat content [m water/DT]
!-------------------------------------------------------------------
qfm=(h-tmp)*rhoice/rhowat
!rtsd  fw=-qfm  !++RT: Why minus??? prec is positive for downward fresh water
         !      water flux too. And besides, salinity of sea ice is
         !      about 5 psu, not zero. In BRIOS we use:
fw=qfm*(34.-5.)/34./dt
         !++RT This is the fresh water flux that should be passed to
	 !     the ocean for coupling.                           -RT--
!-------------------------------------------------------------------
! Heat flux into ML =    !++RT: actually it is the heat flux FROM ML
!  ML heat content left after melting of ice and snow
!  minus ML heat content of previous time step
! Note: In coupled ocean-ice models, the heat content of the
!	mixed layer should be calculated in the ocean part
!-------------------------------------------------------------------
qtm= -tmp2 -(T_oc-TFrez(S_oc))*H_ML*cc/cl

!++RT: This qtm can be used as the ocean surface heat loss.
!++RT: However, I suggest to use ehf further down.

!-------------------------------------------------------------------
! Update snow depth
!-------------------------------------------------------------------
hsn=sn
 
h=max(qhst,0.)             ! New ice thickness
if (h.lt.1E-6) h=0.        ! Avoid very small ice thicknesses
dhgrowth=h-dhgrowth        ! Change in ice thickness due to thermodynamic effects
dhsngrowth=hsn-dhsngrowth  ! Change in snow thickness due to thermodynamic melting
                           ! (without snow fall).
                           ! This is a negative value (= MINUS snow melt)

prec=prec/dt               ! Conversion: 'per time step' -> 'per second'
dhgrowth=dhgrowth/dt       ! Conversion: 'per time step' -> 'per second'
dhsngrowth=dhsngrowth/dt   ! Conversion: 'per time step' -> 'per second'

!RT: This is what I suggest to use as the ocean surface boundary
!RT: condition for coupling:
ehf = ahf + cl*(dhgrowth+(rhosno/rhoice)*dhsngrowth)
!RT FR099:
fw= -dhgrowth*rhoice/rhowat*(34.-5.)/34. - dhsngrowth*rhosno/rhowat + prec
!RT-

!======================================================
! Changes in compactnesses (equation 16 of Hibler 1979)
!======================================================

rh=-min(h,-rh)   ! Make sure we do not try to melt more ice than is available
rA= rhow -(T_oc-TFrez(S_oc))*H_ML*cc/cl

A=A +Saterm*min(rh,0.)*A/max(h,hmin) +max(rA,0.)*(1.-A)/h0   ! Total change. 

A=min(A,h*1.e6)     ! A -> 0 for h -> 0
A=min(max(A,0.),1.) ! A >= 0, A <= 1

!===========================================================
call flooding(h,hsn)     ! Flooding (snow to ice conversion)
!===========================================================

return
end	subroutine therm_ice


!==================================================================

subroutine budget (hice,hsn,t,tair,td,qa,preslev,fsh,flo,ug,S_oc,fh)

!++RT This subroutine needs modification if it is supposed to work with 
!RT   NCEP atmospheric forcing data - which it is.
!RT   I am gonna do it in the final code once it is clear where in the code
!RT   we read the atmospheric daily forcing data.
!RT   I am also gonna do the final crosscheck then.
!--RT

! Thick ice growth rate [m ice/sec]

! INPUT:
! hice - actual ice thickness [m]
! hsn - snow thickness, used for albedo parameterization [m]
! t - temperature of snow/ice surface [C]
! fsh - shortwave radiation [W/m**2]
! flo - longwave radiation  [W/m**2]
! tair - air temperature [C]
! td - dew point air temperature [C]
! preslev - air pressure [hPa]
! ug - wind speed [m/sec]
! S_oc - ocean salinity for the temperature of the ice base calculation [ppt]

! OUTPUT: fh

use ice_therm_parms
implicit none
real*8 hice,hsn,t,Tair,td,Preslev,fsh,flo,             & 
	   fh,ug,hfsen,hfrad,hflat,hftot,S_oc,qa         

real*8  alb             ! Albedo of sea ice
real*8  A1,A2,A3,B,C    ! Local work variables
integer iter, imax      ! Number of iterations

real*8, external :: TFrez

imax = 5

if ((qa == 1.e20).and.(td < 1.e19)) then
 write(*,*)'compute qa from td'
 qa = (0.622/preslev)*6.11*exp(qsi*td/(td+tqi))
endif
 


!-----------------------------------------------------------------
! set albedo
! ice and snow, freezing and melting conditions are distinguished.
!-----------------------------------------------------------------
if (t.lt.0.0) then	        ! freezing condition    ----
     if (hsn.gt.0.0) then	!   snow cover present  -   i
          alb=albsn         !                        i  i
     else              		!   no snow cover       -i  i
          alb=albi       	!                        i  i
     endif		            !   snow cover?         -   
else			            ! melting condition     ----i
     if (hsn.gt.0.0) then	!   snow cover present  -   i
          alb=albsnm	    !                        i  i
     else			        !   no snow cover       -i  i
          alb=albm		    !                        i  i
     endif			        !   snow cover?         -   i
endif			            ! freezing or melting?  ----


! total incoming atmospheric heat flux
a1=(1.0-alb)*fsh +flo +d1*ug*tair +d2i*ug*qa

!
! NEWTON-RHAPSON TO GET TEMPERATURE AT THE TOP OF THE ICE LAYER
!
do iter=1,imax

B=6.11*0.622/preslev*EXP(QSI*T/(T+TQI)) ! specific humidity for ice
A3=D2I*UG*B*QSI*TQI/((T+TQI)**2)        ! gradient coefficient for the humidity part
A2=-D1*UG*T-D2I*UG*B-D3*((T+TMELT)**4)  ! radiation, sensible and latent heat
C=CON/HICE                              ! gradient coefficient for heat conductivity part
A3=A3+4.0*D3*((T+TMELT)**3)+C+D1*UG     ! gradient coefficient for sensible & latent heat 
C=C*(TFrez(S_oc)-T)                        ! downward conductivity term

T=MIN(0.,T+(A1+A2+C)/A3)                ! NEW ICE TEMPERATURE AS THE SUM OF ALL COMPONENTS
end do

!                                         Heat fluxes [W/m**2]
hfrad= (1.0-alb)*fsh &	            	! absorbed short wave radiation
                +flo &              	! long wave radiation coming in
  -d3*((t+tmelt)**4) 	            	! long wave radiation going out

hfsen=d1 *ug*(tair-t)                   ! sensible heat
hflat=d2i*ug*(qa-b)                     ! latent   heat
hftot=hfrad+hfsen+hflat                 ! total (atm.)

fh= -hftot/cl                           ! growth rate [m ice/sec]
!                                      +: ML gains energy, ice melts
!                                      -: ML loses energy, ice grows
return
end subroutine budget

!======================================================

subroutine obudget (preslev,td,qa,fsh,flo,t,ug,tair,fh)  
! Ice growth rate for open ocean [m ice/sec]

! INPUT:
! t - temperature of open water [C]
! fsh - shortwave radiation
! flo - longwave radiation
! tair - air temperature [C]
! td - dew point air temperature [C]
! qa   - specific humidity              !RT
! preslev - air pressure [hPa]
! ug - wind speed [m/sec]

! OUTPUT: fh

!RT: This routine can use either the dew point temperature or the specific humidity.
!RT: Put 1.e20 as a dummy value into the field you do not want to provide.

use ice_therm_parms
implicit none
real*8 t,Tair,preslev,fsh,flo,fh,ug,hfsenow,hfradow,hflatow,hftotow
real*8 td,qa,b

!RT specific humidity [kg/kg] 

!write(*,*)'qaline',qa,td,t,preslev,tqw,qsw

if ((qa >= 9.e19).and.(td < 9.e19)) then
 write (*,*) 'compute specific humidity from dew point temperature'
 qa= (0.622/preslev)*6.11*exp(qsw*td/(td+tqw))
endif

b=6.11*0.622/preslev*exp(qsw*t/(t+tqw)) 

!                                 Heat fluxes [W/m**2]
hfradow= (1.0-albw)*fsh &	            ! absorbed short wave radiation
                   +flo &	            ! long wave radiation coming in
      -d3*((t+tmelt)**4) 	            ! long wave radiation going out

hfsenow=d1 *ug*(tair-t)                 ! sensible heat
hflatow=d2w*ug*(qa-b)                   ! latent   heat
hftotow=hfradow+hfsenow+hflatow ! total

fh= -hftotow/cl                         ! growth rate [m ice/sec]
!                                      +: ML gains energy, ice melts
!                                      -: ML loses energy, ice grows
return
end subroutine obudget

!==========================================================

subroutine flooding (h,hsn)
use ice_therm_parms

real*8 h,hsn,hdraft,hflood

hdraft=(rhosno*hsn+h*rhoice)/rhowat ! Archimedes: displaced water
hflood=hdraft-min(hdraft,h)         ! Increase in mean ice thickness due to flooding
h=h+hflood                          ! Add converted snow to ice volume
hsn=hsn-hflood*rhoice/rhosno        ! Subtract snow from snow layer

!RT    This is what all AWI sea ice models do, but
!RT   I wonder whether it really is correct for the hea budget.
!RT   I suggest we initially keep it to allow for a comparison with BRIOS results
!RT   and rethink it at a later stage.

return
end subroutine flooding

!==========================================================

function TFrez(S)
! Nonlinear correlation for the water freezing temperature.
! Millero (1978) - UNESCO. Reference - See A. Gill, 1982.
implicit none
real*8 S, TFrez

TFrez= -0.0575*S+1.7105e-3 *SQRT(S**3)-2.155e-4 *S*S

!TFrez=-1.86

end function TFrez

!==========================================================

function QToc(T_oc,S_oc)
! Ocean-Ice heat flux.
!RT: not used any more

use ice_therm_parms

implicit none
real*8 T_oc, S_oc, QToc
real*8, external :: TFrez

QToc=rhowat*Cd_thrm_i_o*(T_oc-TFrez(S_oc))

end function QToc
!==========================================================
